# -*- coding: utf-8 -*-
# License: MIT License
# Copyright 2025 InOrbit, Inc.

# Standard
import logging
import logging.config
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from inorbit_connector.models import LoggingConfig


class LogLevels(str, Enum):
    """An enumeration class representing different levels of log messages.

    Log levels for logging.

    See https://docs.python.org/3/library/logging.html#logging-levels

    Attributes:
        DEBUG: Represents the debug log level
        INFO: Represents the info log level
        WARNING: Represents the warning log level
        ERROR: Represents the error log level
        CRITICAL: Represents the critical log level
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def setup_logger(config: "LoggingConfig"):
    """Configures the global logger.

    Args:
        config (LoggingConfig): The logging configuration.
    """
    if config.config_file:
        logging.config.fileConfig(
            config.config_file,
            disable_existing_loggers=False,
            defaults=config.defaults,
        )

    # If a log level is provided, overwrite root logger level
    # This is useful when a log level is set in the YAML file for an specific robot.
    if config.log_level:
        root_logger = logging.getLogger()
        root_logger.setLevel(config.log_level.value)



# -*- coding: utf-8 -*-
# License: MIT License
# Copyright 2025 InOrbit, Inc.

import logging

try:
    import colorlog

    COLORLOG_AVAILABLE = True
except ImportError:
    COLORLOG_AVAILABLE = False


class ConditionalColoredFormatter(logging.Formatter):
    """A formatter that uses colorlog.ColoredFormatter if available,
    otherwise falls back to standard formatting.
    log_colors can be passed in as a dictionary in the defaults section.
    This class is meant to be referenced in logging config files the following way:

    ...
    [formatter_conditional_colored]
    class=inorbit_connector.logging.formatters.ConditionalColoredFormatter
    defaults={'log_colors': {'DEBUG': 'blue','INFO': 'green','WARNING': 'yellow', \
        'ERROR': 'red','CRITICAL': 'bold_red'}}
    ...

    It also allows for the use of colorlog's log_color property.
    """

    def __init__(self, fmt=None, datefmt=None, style="%", defaults=None):
        log_colors = None
        if isinstance(defaults, dict):
            log_colors = defaults.get("log_colors")

        if COLORLOG_AVAILABLE:
            self.formatter = colorlog.ColoredFormatter(
                fmt, datefmt, style, log_colors=log_colors
            )
        else:
            # Remove color codes from format string if colorlog is not available
            fmt = fmt.replace("%(log_color)s", "").replace("%(reset)s", "")
            self.formatter = logging.Formatter(fmt, datefmt, style)

    def format(self, record):
        return self.formatter.format(record)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# License: MIT License
# Copyright 2024 InOrbit, Inc.

# Standard
import os
import logging
from time import sleep
from unittest.mock import AsyncMock, Mock, patch, MagicMock

# Third-party
import pytest
from pydantic import AnyHttpUrl
from inorbit_edge.models import CameraConfig
from inorbit_edge.robot import RobotSession
from pydantic import BaseModel

# InOrbit
from inorbit_connector.connector import (
    Connector,
    InorbitConnectorConfig,
)


class DummyConfig(BaseModel):
    pass


class TestConnectorIsAbstract:
    def test_connector_is_abstract(self):
        assert Connector.__abstractmethods__ == {
            "_connect",
            "_disconnect",
            "_execution_loop",
            "_inorbit_command_handler",
        }

    def test_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            Connector("TestRobot", MagicMock())

    def test_cannot_be_subclassed_without_overriding_abstract_methods(self):
        class SubConnector(Connector):
            pass

        with pytest.raises(TypeError):
            SubConnector("TestRobot", MagicMock())

    def test_can_be_subclassed_with_all_abstract_methods_implemented(self):
        class SubConnector(Connector):
            def _connect(self):
                pass

            def _disconnect(self):
                pass

            def _execution_loop(self):
                pass

            def _inorbit_command_handler(self, command_name, args, options):
                pass

        connector = SubConnector(
            "TestRobot",
            InorbitConnectorConfig(
                api_key="valid_key",
                api_url="https://valid.com/",
                connector_type="valid_connector",
                connector_config=DummyConfig(),
            ),
        )
        assert isinstance(connector, Connector)


class TestConnector:

    @pytest.fixture
    def base_model(self):
        return {
            "api_key": "valid_key",
            "api_url": AnyHttpUrl("https://valid.com/"),
            "connector_type": "valid_connector",
            "connector_config": DummyConfig(),
        }

    @pytest.fixture(autouse=True)
    def make_connector_not_abstract(self):
        Connector.__abstractmethods__ = set()

    @pytest.fixture
    def base_connector(self, base_model):
        return Connector("TestRobot", InorbitConnectorConfig(**base_model))

    def test_init(self, base_model):

        config = InorbitConnectorConfig(**base_model)
        robot_id = "TestRobot"

        connector = Connector(robot_id, config)
        assert connector.robot_id == robot_id
        assert connector.config == config

        assert connector._logger.name == Connector.__module__
        assert logging.getLogger().level == logging.INFO

        assert connector._robot_session.robot_id == robot_id
        assert connector._robot_session.robot_name == robot_id
        assert connector._robot_session.api_key == config.api_key
        assert connector._robot_session.robot_api_key is None
        assert connector._robot_session.endpoint == str(config.api_url)
        assert connector._robot_session.use_ssl is True
        assert connector._robot_session.use_websockets is False
        assert connector._robot_session.robot_key is None

    def test_init_with_robot_key(self, base_model):
        config = InorbitConnectorConfig(
            **base_model, inorbit_robot_key="valid_robot_key"
        )
        robot_id = "TestRobot"

        connector = Connector(robot_id, config)
        assert connector._robot_session.robot_key == "valid_robot_key"

    @pytest.mark.asyncio
    async def test_connect_calls_robot_session_connect(self, base_connector):
        base_connector._robot_session = Mock()
        base_connector._connect = AsyncMock()
        await base_connector._Connector__connect()

        assert base_connector._connect.called
        assert base_connector._robot_session.connect.called

    @pytest.mark.asyncio
    async def test_connect_raises_error_when_failed_to_connect(self, base_connector):
        base_connector._robot_session = Mock()
        base_connector._robot_session.connect.side_effect = Exception(
            "Failed to connect"
        )

        with pytest.raises(Exception) as e:
            await base_connector._Connector__connect()
        assert str(e.value) == "Failed to connect"

    @pytest.mark.asyncio
    async def test_disconnect(self, base_connector):
        base_connector._robot_session = Mock()
        base_connector._disconnect = AsyncMock()
        await base_connector._Connector__disconnect()

        assert base_connector._disconnect.called
        assert base_connector._robot_session.disconnect.called

    def test_start(self, base_model):
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        with patch("threading.Thread") as mock_thread:
            connector._Connector__run_connector = MagicMock()
            connector.start()
            assert not connector._Connector__stop_event.is_set()
            mock_thread.assert_called_with(target=connector._Connector__run_connector)
            mock_thread().start.assert_called_once()

    def test_run_connector(self, base_connector):
        run_connector = base_connector._Connector__run_connector

        with (
            patch("asyncio.new_event_loop") as mock_loop,
            patch("asyncio.set_event_loop") as mock_set_loop,
        ):

            # Setup mocks
            mock_event_loop = MagicMock()
            mock_loop.return_value = mock_event_loop
            base_connector._Connector__connect = AsyncMock()
            base_connector._Connector__run_loop = AsyncMock()
            base_connector._Connector__disconnect = AsyncMock()
            base_connector._robot_session = MagicMock()

            # Call the method
            run_connector()

            # Verify the event loop was created and set
            mock_loop.assert_called_once()
            mock_set_loop.assert_called_once_with(mock_event_loop)

            # Verify connect was called
            base_connector._Connector__connect.assert_called_once()

            # Verify run_loop was called
            base_connector._Connector__run_loop.assert_called_once()

            # Verify disconnect was called in the finally block
            base_connector._Connector__disconnect.assert_called_once()

            # Verify the loop was closed
            mock_event_loop.close.assert_called_once()

    def test_run_with_cameras(self, base_model):
        base_model["cameras"] = [CameraConfig(video_url="https://test.com/")]
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        run_connector = connector._Connector__run_connector

        with (
            patch("asyncio.new_event_loop") as mock_loop,
            patch("asyncio.set_event_loop"),
        ):

            # Setup mocks
            mock_event_loop = MagicMock()
            mock_loop.return_value = mock_event_loop
            connector._Connector__connect = AsyncMock()
            connector._Connector__run_loop = AsyncMock()
            connector._Connector__disconnect = AsyncMock()
            connector._robot_session.register_camera = MagicMock()

            # Call the method
            run_connector()

            assert connector._robot_session.register_camera.call_count == len(
                connector.config.cameras
            )

    def test_run_with_cameras_none_params_ignored(self, base_model):
        base_model["cameras"] = [CameraConfig(video_url="https://test.com/")]
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        run_connector = connector._Connector__run_connector

        with (
            patch("asyncio.new_event_loop") as mock_loop,
            patch("asyncio.set_event_loop"),
        ):

            # Setup mocks
            mock_event_loop = MagicMock()
            mock_loop.return_value = mock_event_loop
            connector._Connector__connect = AsyncMock()
            connector._Connector__run_loop = AsyncMock()
            connector._Connector__disconnect = AsyncMock()

            # Call the method
            run_connector()

            assert len(connector._robot_session.camera_streamers.keys()) == len(
                connector.config.cameras
            )
            assert "0" in connector._robot_session.camera_streamers.keys()
            streamer = connector._robot_session.camera_streamers["0"]
            assert streamer.camera.video_url == str(base_model["cameras"][0].video_url)
            assert streamer.camera.rate == 10
            assert streamer.camera.scaling == 0.3
            assert streamer.camera.quality == 35

    def test_run_with_cameras_custom_params(self, base_model):
        base_model["cameras"] = [
            CameraConfig(video_url="https://test.com/", rate=5, scaling=0.2, quality=30)
        ]
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        run_connector = connector._Connector__run_connector

        with (
            patch("asyncio.new_event_loop") as mock_loop,
            patch("asyncio.set_event_loop"),
        ):

            # Setup mocks
            mock_event_loop = MagicMock()
            mock_loop.return_value = mock_event_loop
            connector._Connector__connect = AsyncMock()
            connector._Connector__run_loop = AsyncMock()
            connector._Connector__disconnect = AsyncMock()

            # Call the method
            run_connector()

            assert len(connector._robot_session.camera_streamers.keys()) == len(
                connector.config.cameras
            )
            assert "0" in connector._robot_session.camera_streamers.keys()
            streamer = connector._robot_session.camera_streamers["0"]
            assert streamer.camera.rate == 5
            assert streamer.camera.scaling == 0.2
            assert streamer.camera.quality == 30
            assert streamer.camera.video_url == str(base_model["cameras"][0].video_url)

    def test_stop(self, base_connector):
        base_connector._Connector__thread = MagicMock()
        base_connector._Connector__thread.is_alive.return_value = False
        assert not base_connector._Connector__stop_event.is_set()
        base_connector.stop()
        assert base_connector._Connector__stop_event.is_set()
        assert base_connector._Connector__thread.join.called

    def test_stop_raises_exception_if_thread_does_not_stop_in_time(
        self, base_connector
    ):
        base_connector._Connector__thread = MagicMock()
        assert not base_connector._Connector__stop_event.is_set()
        with pytest.raises(Exception, match="Thread did not stop in time"):
            base_connector.stop()
            assert base_connector._Connector__stop_event.is_set()
            assert base_connector._Connector__thread.join.called

    def test_run_loop(self, base_model):
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        connector._execution_loop = AsyncMock()
        connector._robot_session = Mock()
        assert not connector._Connector__stop_event.is_set()

        connector.start()
        sleep(1.0 / connector.config.update_freq)
        assert connector._Connector__loop.is_running()
        connector.stop()
        assert not connector._Connector__loop.is_running()

        connector._execution_loop.reset_mock()
        sleep((1.0 / connector.config.update_freq) * 2)
        connector._execution_loop.assert_not_called()

    def test_run_loop_catches_exceptions(self, base_model):
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        connector._execution_loop = AsyncMock(side_effect=Exception("Test exception"))
        connector._robot_session = Mock()
        connector._logger = MagicMock()

        connector.start()
        sleep(1.0 / connector.config.update_freq)
        assert connector._Connector__loop.is_running()
        connector.stop()
        assert not connector._Connector__loop.is_running()
        connector._execution_loop.assert_called()
        connector._logger.error.assert_called()

    def test_publish_map(self, base_model):
        # Test with no maps
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        with patch.object(connector._robot_session, "publish_map") as mock_publish:
            connector.publish_map("map")
            mock_publish.assert_not_called()

        # Test with a map
        base_model["maps"] = {
            "frameA": {
                "file": f"{os.path.dirname(__file__)}/dir/test_map.png",
                "map_id": "valid_map_id",
                "origin_x": 0.0,
                "origin_y": 0.0,
                "resolution": 0.1,
            }
        }
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        with patch.object(connector._robot_session, "publish_map") as mock_publish:
            connector.publish_map("frameA")
            mock_publish.assert_called_once()

    def test_publish_pose_updates_maps(self, base_model):
        base_model["maps"] = {
            "frameA": {
                "file": f"{os.path.dirname(__file__)}/dir/test_map.png",
                "map_id": "valid_map_id",
                "origin_x": 0.0,
                "origin_y": 0.0,
                "resolution": 0.1,
            },
            "frameB": {
                "file": f"{os.path.dirname(__file__)}/dir/test_map.png",
                "map_id": "valid_map_id",
                "origin_x": 0.0,
                "origin_y": 0.0,
                "resolution": 0.1,
            },
        }
        connector = Connector("TestRobot", InorbitConnectorConfig(**base_model))
        with patch.object(connector._robot_session, "publish_map") as mock_publish_map:
            connector.publish_pose(0, 0, 0, "frameA")
            assert mock_publish_map.call_count == 1  # Called on first map publish
            connector.publish_pose(0, 0, 0, "frameA")
            assert mock_publish_map.call_count == 1  # Not called again
            connector.publish_pose(0, 0, 0, "frameB")
            assert mock_publish_map.call_count == 2  # Called again

    def test_register_user_scripts(self, base_model, tmp_path):
        with patch(
            f"{RobotSession.__module__}.{RobotSession.__name__}"
            ".register_commands_path",
            autospec=True,
        ) as mock_register_path:
            # Test it doesn't register callbacks if no user scripts are specified
            Connector("TestRobot", InorbitConnectorConfig(**base_model))
            mock_register_path.assert_not_called()
            mock_register_path.reset_mock()

            # Test it attepts to registers callbacks if user scripts are specified, but
            # fails if the directory does not exist
            Connector(
                "TestRobot",
                InorbitConnectorConfig(**base_model),
                register_user_scripts=True,
                default_user_scripts_dir=tmp_path / "./not_a_dir",
            )
            mock_register_path.assert_not_called()
            mock_register_path.reset_mock()

            # Test it attepts to registers callbacks if user scripts are specified and
            # the directory exists
            Connector(
                "TestRobot",
                InorbitConnectorConfig(**base_model),
                register_user_scripts=True,
                default_user_scripts_dir=tmp_path,
            )
            mock_register_path.assert_called_once()
            mock_register_path.reset_mock()

            # Test it creates the scripts folder if specified
            Connector(
                "TestRobot",
                InorbitConnectorConfig(**base_model),
                register_user_scripts=True,
                default_user_scripts_dir=tmp_path / "a_dir",
                create_user_scripts_dir=True,
            )
            mock_register_path.assert_called_once()
            mock_register_path.reset_mock()

    def test_uses_env_vars(self, base_model):
        base_model["env_vars"] = {"ENV_VAR": "env_value"}
        Connector("TestRobot", InorbitConnectorConfig(**base_model))
        assert "ENV_VAR" in os.environ
        assert os.environ["ENV_VAR"] == "env_value"


class TestConnectorCommandHandler:

    @pytest.fixture
    def base_model(self):
        return {
            "api_key": "valid_key",
            "api_url": AnyHttpUrl("https://valid.com/"),
            "connector_type": "valid_connector",
            "connector_config": DummyConfig(),
        }

    @pytest.fixture(autouse=True)
    def make_connector_not_abstract(self):
        Connector.__abstractmethods__ = set()

    @pytest.fixture
    def base_connector(self, base_model):
        return Connector("TestRobot", InorbitConnectorConfig(**base_model))

    def test_register_command_handler(self, base_model):
        with patch(
            f"{Connector.__module__}.{Connector.__name__}"
            "._register_custom_command_handler",
            autospec=True,
        ) as mock_register_callback:
            # Test it registers by default
            Connector("TestRobot", InorbitConnectorConfig(**base_model))
            mock_register_callback.assert_called_once()
            mock_register_callback.reset_mock()

            # Test it doesn't register if explicitly disabled
            Connector(
                "TestRobot",
                InorbitConnectorConfig(**base_model),
                register_custom_command_handler=False,
            )
            mock_register_callback.assert_not_called()

    def test_handler_wrapper_success(self, base_connector):
        """Test the wrapper function's behavior on successful command execution."""
        connector = base_connector
        connector._robot_session = MagicMock()
        connector._Connector__loop = MagicMock()
        connector._logger = MagicMock()

        with patch("asyncio.run_coroutine_threadsafe") as mock_run_coroutine:

            mock_async_handler = AsyncMock()

            # 1. Register the handler to get the wrapper
            connector._register_custom_command_handler(mock_async_handler)
            wrapper_func = connector._robot_session.register_command_callback.call_args[
                0
            ][0]

            # 2. Prepare arguments for the commands handler
            command_name = "test_command"
            args = ["arg1", "arg2"]
            mock_result_func = MagicMock()
            options = {"result_function": mock_result_func}

            # 3. Call the wrapper
            wrapper_func(command_name, args, options)

            # 4. Assertions
            # Check that run_coroutine_threadsafe was called correctly
            mock_run_coroutine.assert_called_once()
            # Check the coroutine passed to run_coroutine_threadsafe
            assert mock_async_handler.called
            assert mock_async_handler.call_args[0] == (command_name, args, options)
            # Check the loop passed to run_coroutine_threadsafe
            assert mock_run_coroutine.call_args[0][1] is connector._Connector__loop

            # Ensure no error logs were made and result_function wasn't called by the
            # wrapper
            connector._logger.error.assert_not_called()
            mock_result_func.assert_not_called()  # The handler itself should call it

    @pytest.mark.skip(reason="Haven't been able to test the error handling")
    def test_handler_wrapper_exception(self, base_connector):
        """Test wrapper sync handler exception."""
        connector = base_connector
        connector._robot_session = MagicMock()
        connector._Connector__loop = MagicMock()

        # Mock run_coroutine_threadsafe. We only care that it's called.
        with patch("asyncio.run_coroutine_threadsafe") as mock_run_coroutine:

            mock_async_handler = AsyncMock()
            mock_exception = Exception("Test exception")

            # Configure the handler to raise immediately upon call
            mock_async_handler.side_effect = mock_exception

            # 1. Register the handler to get the wrapper
            connector._register_custom_command_handler(mock_async_handler)
            wrapper_func = connector._robot_session.register_command_callback.call_args[
                0
            ][0]

            # 2. Prepare arguments for the wrapper
            cmd_name = "test_command"
            args = ["arg1", "arg2"]
            mock_result_func = MagicMock()
            options = {"result_function": mock_result_func}

            # 3. Call the wrapper
            wrapper_func(cmd_name, args, options)

            # 4. Assertions: Check only the essentials before exception likely halts
            # flow
            # Check run_coroutine_threadsafe was called
            mock_run_coroutine.assert_called_once()
            # Check the handler itself was called (which triggered the side_effect)
            mock_async_handler.assert_called()  # Check it was called at least



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# License: MIT License
# Copyright 2024 InOrbit, Inc.

# Standard
from enum import Enum
import os
import logging
import asyncio
import threading
import traceback
from typing import Coroutine
from abc import ABC, abstractmethod

# Third Party
from inorbit_edge.models import RobotSessionModel
from inorbit_edge.robot import RobotSession
from inorbit_edge.video import OpenCVCamera

# InOrbit
from inorbit_connector.logging.logger import setup_logger
from inorbit_connector.models import InorbitConnectorConfig


class CommandResultCode(str, Enum):
    """The result code of a command execution."""

    SUCCESS = "0"
    FAILURE = "1"


class Connector(ABC):
    """Generic InOrbit connector.

    This is the base class of an InOrbit connector. Subclasses should implement all
    abstract methods.

    A lot of initialization logic is customizable through the configuration object. See
    self.__init__() for more details.
    """

    def __init__(self, robot_id: str, config: InorbitConnectorConfig, **kwargs) -> None:
        """Initialize a new InOrbit connector.

        This class handles bidirectional communication with InOrbit.

        Args:
            robot_id (str): The ID of the InOrbit robot
            config (InorbitConnectorConfig): The connector configuration

        Keyword Args:
            register_user_scripts (bool): Register user scripts automatically.
                Default is False
            default_user_scripts_dir (str): The default user scripts directory path to
                use if not explicitly set in the config.
                Default is "~/.inorbit_connectors/connector-{robot_id}/local/"
            create_user_scripts_dir (bool): The path to the user scripts directory.
                Relevant only if register_user_scripts is True.
                Default is False
        """

        # Common information
        self.robot_id = robot_id
        self.config = config
        self._last_published_frame_id = None

        # Threading for the main run methods
        # The connector runs an asycio loop within a spawned thread
        # self.__loop is initialized within __run_connector(), and only referenced
        # outside of it by the commands handler
        self.__stop_event = asyncio.Event()
        self.__thread = threading.Thread(target=self.__run_loop)
        self.__loop: asyncio.AbstractEventLoop | None = None

        # Logging information
        setup_logger(config.logging)
        self._logger = logging.getLogger(__name__)

        # Set up environment variables
        for env_var_name, env_var_value in config.env_vars.items():
            self._logger.info(f"Setting environment variable '{env_var_name}'")
            os.environ[env_var_name] = env_var_value

        # Create the robot session in InOrbit
        robot_session_config = RobotSessionModel(
            api_key=config.api_key,
            endpoint=config.api_url,
            account_id=config.account_id,
            robot_id=robot_id,
            robot_name=robot_id,
            robot_key=config.inorbit_robot_key,
        )
        self._robot_session = RobotSession(**robot_session_config.model_dump())

        # If enabled, register user scripts
        if kwargs.get("register_user_scripts", False):
            # Get user_scripts path
            path = config.user_scripts_dir
            if path is None:
                path = kwargs.get(
                    "default_user_scripts_dir",
                    f"~/.inorbit_connectors/connector-{robot_id}/local/",
                )
            user_scripts_path = os.path.expanduser(path)
            create_dir = kwargs.get("create_user_scripts_dir", False)
            self._register_user_scripts(user_scripts_path, create_dir)

        # If enabled, register the provided custom commands handler
        if kwargs.get("register_custom_command_handler", True):
            self._register_custom_command_handler(self._inorbit_command_handler)

    def _register_user_scripts(self, path: str, create: bool) -> None:
        """Register user scripts folder.

        Args:
            path (str): The path to the user scripts directory.
            create (bool): Create the directory if it doesn't exist.
        """
        if not os.path.exists(path):
            if create:
                self._logger.info(f"Creating user_scripts directory: {path}")
                os.makedirs(path, exist_ok=True)
            else:
                self._logger.warning(f"User_scripts directory not found: {path}")
                return
        if os.path.exists(path):
            self._logger.info(f"Registering user_scripts path: {path}")
            # NOTE: this only supports bash execution (exec_name_regex is set to
            # files with '.sh' extension).
            # More script types can be supported, but right now is only limited to
            # bash scripts
            self._robot_session.register_commands_path(path, exec_name_regex=r".*\.sh")

    def _register_custom_command_handler(self, async_handler: Coroutine) -> None:
        """Register an async custom command handler wrapped in error handling logic.

        Args:
            async_handler (Coroutine): The custom commands handler.
        """

        def handler_wrapper(command_name: str, args: list, options: dict):
            try:
                # Handle the commands in the existing event loop and wait for the result
                asyncio.run_coroutine_threadsafe(
                    async_handler(command_name, args, options), self.__loop
                ).result()
            except Exception as e:
                self._logger.error(
                    f"Failed to execute command '{command_name}' with args {args}. "
                    f"Exception:\n{str(e) or e.__class__.__name__}"
                )
                options["result_function"](
                    CommandResultCode.FAILURE,
                    execution_status_details=(
                        "An error occured executing custom command"
                    ),
                    stderr=str(e) or e.__class__.__name__,
                )

        self._robot_session.register_command_callback(handler_wrapper)

    @abstractmethod
    async def _inorbit_command_handler(
        self, command_name: str, args: list, options: dict
    ):
        """Callback method for command messages. This method is called when a command
        is received from InOrbit.
        Will automatically be registered if `register_custom_command_handler`
        constructor keyword argument is set, which is the default behavior.

        The result function will always be included in the options dictionary and must
        be called in order to report the result of a command. It has the following
        signature:

        options['result_function'](
            result_code: CommandResultCode,
            execution_status_details: str | None = None,
            stdout: str | None = None,
            stderr: str | None = None,
        ) -> None

        e.g.:
        if success:
            return options['result_function'](CommandResultCode.SUCCESS)
        else:
            return options['result_function'](
                CommandResultCode.FAILURE,
                stderr="Example error"
            )

        Args:
            command_name (str): The name of the command
            args (list): The list of arguments
            options (dict): The dictionary of options.
                It contains the `result_function` explained above.
        """
        pass

    @abstractmethod
    async def _connect(self) -> None:
        """Connect to any external services.

        This method should not be called directly. Instead, call the start() method to
        start the connector. This ensures that the connector is only started once.
        """
        pass

    async def __connect(self) -> None:
        """Initialize the connection to InOrbit based on the provided configuration,
        and connect to external services calling self._connect().

        Raises:
            Exception: If the robot session cannot connect.
        """
        # Call the user-implemented connection logic
        await self._connect()

        # Connect to InOrbit
        self._robot_session.connect()

    @abstractmethod
    async def _disconnect(self) -> None:
        """Disconnect from any external services.

        This method should not be called directly. Instead, call the stop() method to
        stop the connector. This ensures that the connector is only stopped once.
        """
        pass

    async def __disconnect(self) -> None:
        """Disconnect external services and disconnect from InOrbit."""

        # Disconnect from InOrbit
        self._robot_session.disconnect()

        # Call the user-implemented disconnection logic
        await self._disconnect()

    @abstractmethod
    async def _execution_loop(self) -> None:
        """The main execution loop for the connector.

        This method should be overridden by subclasses to provide the execution loop for
        the connector, will be called repeatedly until the connector is stopped, and
        should not be called directly. Instead, call the start() or stop() methods to
        start or stop the connector. This ensures that the connector is only started or
        stopped once.
        """
        pass

    def publish_map(self, frame_id: str, is_update: bool = False) -> None:
        """Publish the map metadata to InOrbit. If `frame_id` is not found in the maps
        configuration, this method will not publish anything.
        """
        if map_config := self.config.maps.get(frame_id):
            self._robot_session.publish_map(
                file=map_config.file,
                map_id=map_config.map_id,
                frame_id=frame_id,
                x=map_config.origin_x,
                y=map_config.origin_y,
                resolution=map_config.resolution,
                ts=None,
                is_update=is_update,
            )
            self._last_published_frame_id = frame_id
        else:
            self._logger.error(
                f"Map {frame_id} not found in the current configuration."
                " Map message will not be sent."
            )

    def publish_pose(
        self, x: float, y: float, yaw: float, frame_id: str, *args, **kwargs
    ) -> None:
        """Publish a pose to InOrbit. If the frame_id is different from the last
        published, it calls self.publish_map() to update the map.
        """
        if frame_id != self._last_published_frame_id:
            self._logger.info(f"Updating map {frame_id} with new pose.")
            self.publish_map(frame_id, is_update=True)
        self._robot_session.publish_pose(x, y, yaw, frame_id, *args, **kwargs)

    def start(self) -> None:
        """Start the connector in a new thread.

        This method should be called to start the execution of this connector. It
        creates an event loop in a new thread and runs the connector in it.

        After calling start(), use join() to block until the connector is stopped.
        Use stop() to stop the connector.

        It:
        - calls self._connect() to connect to any external services.
        - sets up camera feeds defined in the configuration.
        - runs the execution loop in a new thread.
        - calls self._disconnect() to disconnect from any external services once the
          connector is stopped.
        """

        # Prevent starting an already running thread
        if not self.__thread.is_alive():
            self.__stop_event.clear()

            # Create and start the thread
            self.__thread = threading.Thread(target=self.__run_connector)
            self.__thread.start()

    def join(self) -> None:
        """Join the execution loop of this connector.

        This method should be called to join the execution loop of this connector and
        will block until it ends.
        """
        self.__thread.join()

    def stop(self) -> None:
        """Stop the execution loop of this connector.

        This method should be called to stop the execution loop of this connector, will
        block until the execution loop is stopped, and will call disconnect() to clean
        up any external connections.
        """

        # Stop the execution loop
        self._logger.info("Stopping connector")
        self.__stop_event.set()
        self.__thread.join(timeout=5)
        if self.__thread.is_alive():
            raise Exception("Thread did not stop in time")

    def __run_connector(self):
        """The target function of the connector's thread.

        It connects to InOrbit via the edge-sdk, start the execution loop and
        disconnects all services when the connector is signaled stop via self.stop().
        """

        # Create a new event loop for this thread
        self.__loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.__loop)

        # Connect to external services and create the InOrbit session
        self.__loop.run_until_complete(self.__connect())

        # Set up camera feeds
        for idx, camera_config in enumerate(self.config.cameras):
            self._logger.info(
                f"Registering camera {idx}: {str(camera_config.video_url)}"
            )
            # If values are None, use default instead
            dump = camera_config.model_dump()
            clean = {k: v for k, v in dump.items() if v is not None}
            self._robot_session.register_camera(str(idx), OpenCVCamera(**clean))

        try:
            self.__loop.run_until_complete(self.__run_loop())
        except Exception as e:
            self._logger.error(f"Error in execution loop: {e}")
            self._logger.error(f"Traceback: {traceback.format_exc()}")
        finally:
            self.__loop.run_until_complete(self.__disconnect())
            self.__loop.close()

    async def __run_loop(self) -> None:
        """The main coroutine of the connector.

        This coroutine will run the execution loop of the connector until the stop event
        is set.
        It uses self.config.update_freq to set a maximum frequency for the execution
        loop, but it will never run faster than the actual execution of the loop body.

        Exceptions raised by self._execution_loop() are caught and logged to prevent
        the connector from crashing. It is recommended to handle exceptions within the
        loop and publish the errors.
        """
        while not self.__stop_event.is_set():
            try:
                await asyncio.gather(
                    self._execution_loop(),
                    asyncio.sleep(1.0 / self.config.update_freq),
                )
            except Exception as e:
                self._logger.error(f"Error in execution loop: {e}")
                self._logger.error(f"Traceback: {traceback.format_exc()}")
                # Continue execution after a brief pause to avoid tight error loops
                await asyncio.sleep(1.0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# License: MIT License
# Copyright 2025 InOrbit, Inc.

# Third-party
import pytest

# InOrbit
from inorbit_connector.logging.logger import LogLevels
from inorbit_connector.models import LoggingConfig


class TestLoggingConfig:
    def test_default_values(self):
        config = LoggingConfig()
        assert config.config_file is not None
        assert config.log_level is None
        assert config.defaults == {
            "log_file": "inorbit-connector.log",
        }

    def test_custom_values(self):
        config = LoggingConfig(
            log_level=LogLevels.INFO, defaults={"log_file": "./test.log"}
        )
        assert config.log_level == LogLevels.INFO
        assert config.defaults == {"log_file": "./test.log"}

    def test_invalid_log_level(self):
        with pytest.raises(
            ValueError,
            match="Input should be 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'",
        ):
            LoggingConfig(log_level="BAD")
